# -*- coding: utf-8 -*-
"""
gdcdatamodel.gdc_postgres_admin
----------------------------------

Module for stateful management of a GDC PostgreSQL installation.
"""

import argparse
import logging
import random
import sqlalchemy as sa
import time

from collections import namedtuple
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError

#: Required but 'unused' import to register GDC models
from . import models  # noqa

from psqlgraph import (
    create_all,
    Node,
    Edge,
)

logging.basicConfig()
logger = logging.getLogger("gdc_postgres_admin")
logger.setLevel(logging.INFO)

name_root = "table_creator_"
app_name = "{}{}".format(name_root, random.randint(1000, 9999))
no_kill_list = []
BlockingQueryResult = namedtuple(
    "BlockingQueryResult",
    [
        "blocked_appname",
        "blocked_pid",
        "blocking_appname",
        "blocking_pid",
        "blocking_statement",
    ],
)


# See https://wiki.postgresql.org/wiki/Lock_Monitoring
BLOCKING_SQL = """

SELECT
    blocked_activity.application_name  AS blocked_appname,
    blocked_locks.pid                  AS blocked_pid,

    blocking_activity.application_name AS blocking_appname,
    blocking_locks.pid                 AS blocking_pid,

    blocking_activity.query            AS blocking_statement

FROM pg_catalog.pg_locks               blocked_locks

JOIN pg_catalog.pg_stat_activity       blocked_activity
    ON blocked_activity.pid            = blocked_locks.pid

JOIN pg_catalog.pg_locks               blocking_locks
    ON  blocking_locks.locktype        = blocked_locks.locktype
    AND blocking_locks.DATABASE        IS NOT DISTINCT FROM blocked_locks.DATABASE
    AND blocking_locks.relation        IS NOT DISTINCT FROM blocked_locks.relation
    AND blocking_locks.page            IS NOT DISTINCT FROM blocked_locks.page
    AND blocking_locks.tuple           IS NOT DISTINCT FROM blocked_locks.tuple
    AND blocking_locks.virtualxid      IS NOT DISTINCT FROM blocked_locks.virtualxid
    AND blocking_locks.transactionid   IS NOT DISTINCT FROM blocked_locks.transactionid
    AND blocking_locks.classid         IS NOT DISTINCT FROM blocked_locks.classid
    AND blocking_locks.objid           IS NOT DISTINCT FROM blocked_locks.objid
    AND blocking_locks.objsubid        IS NOT DISTINCT FROM blocked_locks.objsubid
    AND blocking_locks.pid             != blocked_locks.pid

JOIN pg_catalog.pg_stat_activity blocking_activity
     ON blocking_activity.pid          = blocking_locks.pid

WHERE NOT blocked_locks.GRANTED
      AND blocked_activity.application_name = :app_name;

"""


GRANT_READ_PRIVS_SQL = """
BEGIN;
GRANT SELECT ON TABLE {table} TO {user};
COMMIT;
"""

GRANT_WRITE_PRIVS_SQL = """
BEGIN;
GRANT SELECT, INSERT, UPDATE, DELETE ON TABLE {table} TO {user};
COMMIT;
"""

REVOKE_READ_PRIVS_SQL = """
BEGIN;
REVOKE SELECT, INSERT, UPDATE, DELETE ON TABLE {table} FROM {user};
COMMIT;
"""

REVOKE_WRITE_PRIVS_SQL = """
BEGIN;
REVOKE INSERT, UPDATE, DELETE ON TABLE {table} FROM {user};
COMMIT;
"""


def execute(engine, sql, *args, **kwargs):
    statement = sa.sql.text(sql)
    logger.debug(statement)
    return engine.execute(statement, *args, **kwargs)


def get_engine(host, user, password, database):
    connect_args = {"application_name": app_name}
    con_str = "postgres://{user}:{pwd}@{host}/{db}".format(
        user=user, host=host, pwd=password, db=database
    )
    return create_engine(con_str, connect_args=connect_args)


def execute_for_all_graph_tables(engine, sql, *args, **kwargs):
    """Execute a SQL statment that has a python format variable {table}
    to be replaced with the tablename for all Node and Edge tables

    """
    for cls in Node.__subclasses__() + Edge.__subclasses__():
        _kwargs = dict(kwargs, **{"table": cls.__tablename__})
        statement = sql.format(**_kwargs)
        execute(engine, statement)


def grant_read_permissions_to_graph(engine, user):
    execute_for_all_graph_tables(engine, GRANT_READ_PRIVS_SQL, user=user)


def grant_write_permissions_to_graph(engine, user):
    execute_for_all_graph_tables(engine, GRANT_WRITE_PRIVS_SQL, user=user)


def revoke_read_permissions_to_graph(engine, user):
    execute_for_all_graph_tables(engine, REVOKE_READ_PRIVS_SQL, user=user)


def revoke_write_permissions_to_graph(engine, user):
    execute_for_all_graph_tables(engine, REVOKE_WRITE_PRIVS_SQL, user=user)


def migrate_transaction_snapshots(engine, user):
    """
    Updates to TransactionSnapshot table:
        - change old `id` column to `entity_id`, which is no longer unique or primary
          key
        - add new serial `id` column as primary key
    """
    md = MetaData(bind=engine)
    tablename = models.submission.TransactionSnapshot.__tablename__
    snapshots_table = Table(tablename, md, autoload=True)
    if "entity_id" not in snapshots_table.c:
        execute(
            engine,
            "ALTER TABLE {name} DROP CONSTRAINT {name}_pkey".format(name=tablename),
        )
        execute(
            engine,
            "ALTER TABLE {} RENAME id TO entity_id".format(tablename),
        )
        execute(
            engine,
            "ALTER TABLE {} ADD COLUMN id SERIAL PRIMARY KEY;".format(tablename),
        )


def create_graph_tables(engine, timeout):
    """
    create a table
    """
    logger.info("Creating tables (timeout: %d)", timeout)

    connection = engine.connect()
    trans = connection.begin()
    logger.info("Setting lock_timeout to %d", timeout)

    timeout_str = "{}s".format(int(timeout + 1))
    connection.execute("SET LOCAL lock_timeout = %s;", timeout_str)

    create_all(connection)
    trans.commit()


def is_blocked_by_no_kill(blocking):
    for proc in blocking:
        if proc.blocking_appname in no_kill_list:
            print(
                "Blocked by no-kill process {}, {}: {}".format(
                    proc.blocking_appname, proc.blocking_pid, proc.blocking_statement
                )
            )
            return True
    return False


def lookup_blocking_psql_backend_processes(engine):
    """"""

    sql_cmd = sa.sql.text(BLOCKING_SQL)
    conn = engine.connect()
    blocking = conn.execute(sql_cmd, app_name=app_name)
    return [BlockingQueryResult(*b) for b in blocking]


def kill_blocking_psql_backend_processes(engine):
    """Query the postgres backend tables for the process that is blocking
    this app, as identified by the `app_name`.

    .. warning:: **THIS COMMAND KILLS OTHER PEOPLES POSTGRES QUERIES.**

    It is sometimes necessary to kill other peoples queries in order
    to gain a write lock on a table to ALTER it for a foreign-key from
    a new table.

    There is a list at the top of this module that specifies which
    processes are 'no-kill'.  There currently are none, but a good
    exmaple of one that you might want to put in there is the
    Elasticsearch build process, since you might not want to kill a 5h
    long process 4h in.

    """

    blockers = lookup_blocking_psql_backend_processes(engine)

    if is_blocked_by_no_kill(blockers):
        logger.warn("Process blocked by a 'no-kill' process. " "Refusing to kill it")
        return

    if not blockers:
        logger.warning("Found %d blocking processes!", len(blockers))
    else:
        logger.info("Found %d blocking processes", len(blockers))

    for result in blockers:
        logger.warning(
            "Killing blocking backend process: name({})\tpid({}): {}".format(
                result.blocking_appname, result.blocking_pid, result.blocking_statement
            )
        )

        # Kill anything in the way, it was deemed of low importance
        sql_cmd = "SELECT pg_terminate_backend({blocking_pid});".format(
            blocking_pid=result.blocking_pid
        )
        execute(engine, sql_cmd)


def create_tables_force(engine, delay, retries):
    """Create the tables and **KILL ANY BLOCKING PROCESSES**.

    This command will spawn a process to create the new tables in
    order to find out which process is blocking us.  If we didn't do
    this concurrently, then the table creation will have disappeared
    by the time we tried to find its blocker in the postgres backend
    tables.

    """

    logger.info("Running table creator named %s", app_name)
    logger.warning("Running with force=True option %s", app_name)

    from multiprocessing import Process

    p = Process(target=create_graph_tables, args=(engine, delay))
    p.start()
    time.sleep(delay)

    if p.is_alive():
        logger.warning("Table creation blocked!")
        kill_blocking_psql_backend_processes(engine)

        #  Wait some time for table creation to proceed
        time.sleep(4)

    if p.is_alive():
        if retries <= 0:
            raise RuntimeError("Max retries exceeded.")

        logger.warning("Table creation failed, retrying.")
        return create_tables_force(engine, delay, retries - 1)


def create_tables(engine, delay, retries):
    """Create the tables but do not kill any blocking processes.

    This command will catch OperationalErrors signalling timeouts from
    the database when the lock was not obtained successfully within
    the `delay` period.

    """

    logger.info("Running table creator named %s", app_name)
    try:
        return create_graph_tables(engine, delay)

    except OperationalError as e:
        if "timeout" in str(e):
            logger.warning("Attempt timed out")
        else:
            raise

        if retries <= 0:
            raise RuntimeError("Max retries exceeded")

        logger.info(
            "Trying again in {} seconds ({} retries remaining)".format(delay, retries)
        )
        time.sleep(delay)

        create_tables(engine, delay, retries - 1)


def subcommand_create(args):
    """Idempotently/safely create ALL tables in database that are required
    for the GDC.  This command will not delete/drop any data.

    """

    logger.info("Running subcommand 'create'")
    engine = get_engine(args.host, args.user, args.password, args.database)
    kwargs = dict(
        engine=engine,
        delay=args.delay,
        retries=args.retries,
    )

    if args.force:
        return create_tables_force(**kwargs)
    else:
        return create_tables(**kwargs)


def subcommand_grant(args):
    """Grant permissions to a user.

    Argument ``--read`` will grant users read permissions
    Argument ``--write`` will grant users write and READ permissions
    """

    logger.info("Running subcommand 'grant'")
    engine = get_engine(args.host, args.user, args.password, args.database)

    assert args.read or args.write, "No premission types/users specified."

    if args.read:
        users_read = [u for u in args.read.split(",") if u]
        for user in users_read:
            grant_read_permissions_to_graph(engine, user)

    if args.write:
        users_write = [u for u in args.write.split(",") if u]
        for user in users_write:
            grant_write_permissions_to_graph(engine, user)


def subcommand_revoke(args):
    """Grant permissions to a user.

    Argument ``--read`` will revoke users' read permissions
    Argument ``--write`` will revoke users' write AND READ permissions
    """

    logger.info("Running subcommand 'revoke'")
    engine = get_engine(args.host, args.user, args.password, args.database)

    if args.read:
        users_read = [u for u in args.read.split(",") if u]
        for user in users_read:
            revoke_read_permissions_to_graph(engine, user)

    if args.write:
        users_write = [u for u in args.write.split(",") if u]
        for user in users_write:
            revoke_write_permissions_to_graph(engine, user)


def add_base_args(subparser):
    subparser.add_argument(
        "-H", "--host", type=str, action="store", required=True, help="psql-server host"
    )
    subparser.add_argument(
        "-U", "--user", type=str, action="store", required=True, help="psql test user"
    )
    subparser.add_argument(
        "-D",
        "--database",
        type=str,
        action="store",
        required=True,
        help="psql test database",
    )
    subparser.add_argument(
        "-P",
        "--password",
        type=str,
        action="store",
        default="",
        help="psql test password",
    )
    return subparser


def add_subcommand_create(subparsers):
    parser = add_base_args(
        subparsers.add_parser("graph-create", help=subcommand_create.__doc__)
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Hard killing blocking processes that are not in the 'no-kill' list.",
    )
    parser.add_argument(
        "--delay",
        type=int,
        action="store",
        default=60,
        help="How many seconds to wait for blocking processes to finish before retrying (and hard killing them if used with --force).",
    )
    parser.add_argument(
        "--retries",
        type=int,
        action="store",
        default=10,
        help="If blocked by important process, how many times to retry after waiting `delay` seconds.",
    )


def add_subcommand_grant(subparsers):
    parser = add_base_args(
        subparsers.add_parser("graph-grant", help=subcommand_grant.__doc__)
    )
    parser.add_argument(
        "--read",
        type=str,
        action="store",
        help="Users to grant read access to (comma separated).",
    )
    parser.add_argument(
        "--write",
        type=str,
        action="store",
        help="Users to grant read/write access to (comma separated).",
    )


def add_subcommand_revoke(subparsers):
    parser = add_base_args(
        subparsers.add_parser("graph-revoke", help=subcommand_revoke.__doc__)
    )
    parser.add_argument(
        "--read",
        type=str,
        action="store",
        help="Users to revoke read access from (comma separated).",
    )
    parser.add_argument(
        "--write",
        type=str,
        action="store",
        help=(
            "Users to revoke write access from (comma separated). "
            "NOTE: The user will still have read privs!!"
        ),
    )


def get_parser():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subcommand")
    add_subcommand_create(subparsers)
    add_subcommand_grant(subparsers)
    add_subcommand_revoke(subparsers)
    return parser


def main(args=None):
    args = args or get_parser().parse_args()

    logger.info("[ HOST     : %-10s ]", args.host)
    logger.info("[ DATABASE : %-10s ]", args.database)
    logger.info("[ USER     : %-10s ]", args.user)

    return_value = {
        "graph-create": subcommand_create,
        "graph-grant": subcommand_grant,
        "graph-revoke": subcommand_revoke,
    }[args.subcommand](args)

    logger.info("Done.")
    return return_value



from graphviz import Digraph


def create_graphviz(nodes, include_case_cache_edges=False):
    """Returns Graphviz Dot object.

    Call dot.render(path) to write to file
    """

    dot = Digraph()
    dot.graph_attr["rankdir"] = "RL"

    edges_added = set()
    nodes = {node.node_id: node for node in nodes}

    def is_edge_drawn(edge, neighbor):
        is_case_cache_edge = "RelatesToCase" in edge.__class__.__name__

        return (
            (include_case_cache_edges or not is_case_cache_edge)
            and edge not in edges_added
            and neighbor in nodes
        )

    for node in nodes.values():
        dot.node(node.node_id, str(node))

        for edge in node.edges_out:
            if is_edge_drawn(edge, edge.dst_id):
                dot.edge(edge.src_id, edge.dst_id)
                edges_added.add(edge)

        for edge in node.edges_in:
            if is_edge_drawn(edge, edge.src_id):
                dot.edge(edge.src_id, edge.dst_id)
                edges_added.add(edge)

    return dot



import unittest
import logging
from psqlgraph.exc import ValidationError
from gdcdatamodel import models as md

logging.basicConfig(level=logging.INFO)


class TestDataModel(unittest.TestCase):
    def test_type_validation(self):
        f = md.File()
        with self.assertRaises(ValidationError):
            f.file_size = "0"
        f.file_size = 0

        f = md.File()
        with self.assertRaises(ValidationError):
            f.file_name = 0
        f.file_name = "0"

        s = md.Sample()
        with self.assertRaises(ValidationError):
            s.is_ffpe = "false"
        s.is_ffpe = False

        s = md.Slide()
        with self.assertRaises(ValidationError):
            s.percent_necrosis = "0.0"
        s.percent_necrosis = 0.0

    def test_link_clobber_prevention(self):
        with self.assertRaises(AssertionError):
            md.EdgeFactory(
                "Testedge",
                "test",
                "sample",
                "aliquot",
                "aliquots",
                "_uncontended_backref",
            )

    def test_backref_clobber_prevention(self):
        with self.assertRaises(AssertionError):
            md.EdgeFactory(
                "Testedge",
                "test",
                "sample",
                "aliquot",
                "_uncontended_link",
                "samples",
            )



from sqlalchemy import Column, Integer, String, Text, DateTime, text, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql import ARRAY

import json

Base = declarative_base()


class Notification(Base):
    __tablename__ = "notifications"

    id = Column(Integer, primary_key=True)
    components = Column(ARRAY(Text), default=list())
    message = Column(String)
    level = Column(String)
    dismissible = Column(Boolean, default=True)
    created = Column(
        DateTime(timezone=True),
        nullable=False,
        server_default=text("now()"),
    )

    def __repr__(self):
        return "<Notification(id='{}', level='{}', message='{}')>".format(
            self.id, self.level, self.message
        )

    def to_json(self):
        """Returns a JSON safe representation of :class:`Notification`"""

        return json.loads(
            json.dumps(
                {
                    "id": self.id,
                    "components": self.components,
                    "dismissible": self.dismissible,
                    "message": self.message,
                    "level": self.level,
                }
            )
        )



# -*- coding: utf-8 -*-
"""
Tests for gdcdatamodel.gdc_postgres_admin module
"""

import logging
import unittest

from gdcdatamodel import gdc_postgres_admin as pgadmin
from gdcdatamodel import models
from sqlalchemy.exc import ProgrammingError

from multiprocessing import (
    Process,
    Queue,
)

from psqlgraph import (
    Edge,
    Node,
    PsqlGraphDriver,
)

logging.basicConfig()


class TestGDCPostgresAdmin(unittest.TestCase):

    logger = logging.getLogger("TestGDCPostgresAdmin")
    logger.setLevel(logging.INFO)

    host = "localhost"
    user = "postgres"
    database = "automated_test"

    base_args = [
        "-H",
        host,
        "-U",
        user,
        "-D",
        database,
    ]

    g = PsqlGraphDriver(host, user, "", database)
    root_con_str = "postgres://{user}:{pwd}@{host}/{db}".format(
        user=user, host=host, pwd="", db=database
    )
    engine = pgadmin.create_engine(root_con_str)

    @classmethod
    def tearDownClass(cls):
        """Recreate the database for tests that follow."""
        cls.create_all_tables()

        # Re-grant permissions to test user
        for scls in Node.__subclasses__() + Edge.__subclasses__():
            statment = "GRANT ALL PRIVILEGES ON TABLE {} TO test".format(
                scls.__tablename__
            )
            cls.engine.execute("BEGIN; %s; COMMIT;" % statment)

    @classmethod
    def drop_all_tables(cls):
        for scls in Node.__subclasses__():
            try:
                cls.engine.execute("DROP TABLE {} CASCADE".format(scls.__tablename__))
            except Exception as e:
                cls.logger.warning(e)

    @classmethod
    def create_all_tables(cls):
        parser = pgadmin.get_parser()
        args = parser.parse_args(
            ["graph-create", "--delay", "1", "--retries", "0", "--force"]
            + cls.base_args
        )
        pgadmin.main(args)

    @classmethod
    def drop_a_table(cls):
        cls.engine.execute("DROP TABLE edge_clinicaldescribescase")
        cls.engine.execute("DROP TABLE node_clinical")

    def startTestRun(self):
        self.drop_all_tables()

    def setUp(self):
        self.drop_all_tables()

    def test_args(self):
        parser = pgadmin.get_parser()
        parser.parse_args(["graph-create"] + self.base_args)

    def test_create_single(self):
        """Test simple table creation"""

        pgadmin.main(
            pgadmin.get_parser().parse_args(
                ["graph-create", "--delay", "1", "--retries", "0"] + self.base_args
            )
        )

        self.engine.execute("SELECT * from node_case")

    def test_create_double(self):
        """Test idempotency of table creation"""

        pgadmin.main(
            pgadmin.get_parser().parse_args(
                ["graph-create", "--delay", "1", "--retries", "0"] + self.base_args
            )
        )

        self.engine.execute("SELECT * from node_case")

    def test_create_fails_blocked_without_force(self):
        """Test table creation fails when blocked w/o force"""

        q = Queue()  # to communicate with blocking process

        args = pgadmin.get_parser().parse_args(
            ["graph-create", "--delay", "1", "--retries", "1"] + self.base_args
        )
        pgadmin.main(args)

        self.drop_a_table()

        def blocker():
            with self.g.session_scope() as s:
                s.merge(models.Case("1"))
                q.put(0)  # Tell main thread we're ready
                q.get()  # Wait for main thread to tell us to exit

        p = Process(target=blocker)
        p.daemon = True
        p.start()
        q.get()

        with self.assertRaises(RuntimeError):
            pgadmin.main(args)

        q.put(0)
        p.terminate()

    def test_create_force(self):
        """Test ability to force table creation"""

        q = Queue()  # to communicate with blocking process

        args = pgadmin.get_parser().parse_args(
            ["graph-create", "--delay", "1", "--retries", "1", "--force"]
            + self.base_args
        )
        pgadmin.main(args)

        self.drop_a_table()

        def blocker():
            with self.g.session_scope() as s:
                s.merge(models.Case("1"))
                q.put(0)  # Tell main thread we're ready
                q.get()  # This get should block until this prcoess is killed
                assert False, "Should not be reachable!"

        p = Process(target=blocker)
        p.daemon = True
        p.start()
        q.get()

        try:
            pgadmin.main(args)
        except:
            p.terminate()
            raise

        q.put(0)
        p.terminate()

    def test_priv_grant_read(self):
        """Test ability to grant read but not write privs"""

        self.create_all_tables()
        self.engine.execute("CREATE USER pytest WITH PASSWORD 'pyt3st'")

        try:
            g = PsqlGraphDriver(self.host, "pytest", "pyt3st", self.database)

            #: If this failes, this test (not the code) is wrong!
            with self.assertRaises(ProgrammingError):
                with g.session_scope():
                    g.nodes().count()

            pgadmin.main(
                pgadmin.get_parser().parse_args(
                    [
                        "graph-grant",
                        "--read=pytest",
                    ]
                    + self.base_args
                )
            )

            with g.session_scope():
                g.nodes().count()

            with self.assertRaises(ProgrammingError):
                with g.session_scope() as s:
                    s.merge(models.Case("1"))

        finally:
            self.engine.execute("DROP OWNED BY pytest; DROP USER pytest")

    def test_priv_grant_write(self):
        """Test ability to grant read/write privs"""

        self.create_all_tables()
        self.engine.execute("CREATE USER pytest WITH PASSWORD 'pyt3st'")

        try:
            g = PsqlGraphDriver(self.host, "pytest", "pyt3st", self.database)
            pgadmin.main(
                pgadmin.get_parser().parse_args(
                    [
                        "graph-grant",
                        "--write=pytest",
                    ]
                    + self.base_args
                )
            )

            with g.session_scope() as s:
                g.nodes().count()
                s.merge(models.Case("1"))

        finally:
            self.engine.execute("DROP OWNED BY pytest; DROP USER pytest")

    def test_priv_revoke_read(self):
        """Test ability to revoke read privs"""

        self.create_all_tables()
        self.engine.execute("CREATE USER pytest WITH PASSWORD 'pyt3st'")

        try:
            g = PsqlGraphDriver(self.host, "pytest", "pyt3st", self.database)

            pgadmin.main(
                pgadmin.get_parser().parse_args(
                    [
                        "graph-grant",
                        "--read=pytest",
                    ]
                    + self.base_args
                )
            )

            pgadmin.main(
                pgadmin.get_parser().parse_args(
                    [
                        "graph-revoke",
                        "--read=pytest",
                    ]
                    + self.base_args
                )
            )

            with self.assertRaises(ProgrammingError):
                with g.session_scope() as s:
                    g.nodes().count()
                    s.merge(models.Case("1"))

        finally:
            self.engine.execute("DROP OWNED BY pytest; DROP USER pytest")

    def test_priv_revoke_write(self):
        """Test ability to revoke read/write privs"""

        self.create_all_tables()
        self.engine.execute("CREATE USER pytest WITH PASSWORD 'pyt3st'")

        try:
            g = PsqlGraphDriver(self.host, "pytest", "pyt3st", self.database)

            pgadmin.main(
                pgadmin.get_parser().parse_args(
                    [
                        "graph-grant",
                        "--write=pytest",
                    ]
                    + self.base_args
                )
            )

            pgadmin.main(
                pgadmin.get_parser().parse_args(
                    [
                        "graph-revoke",
                        "--write=pytest",
                    ]
                    + self.base_args
                )
            )

            with g.session_scope() as s:
                g.nodes().count()

            with self.assertRaises(ProgrammingError):
                with g.session_scope() as s:
                    s.merge(models.Case("1"))

        finally:
            self.engine.execute("DROP OWNED BY pytest; DROP USER pytest")



# -*- coding: utf-8 -*-
"""
gdcdatamodel.test.conftest
----------------------------------

Test GDC specific index creation.

"""


def test_secondary_key_indexes(indexes):
    assert "index_node_datasubtype_name_lower" in indexes
    assert "index_node_analyte_project_id" in indexes
    assert "index_4df72441_famihist_submitte_id_lower" in indexes
    assert "transaction_logs_project_id_idx" in indexes



# -*- coding: utf-8 -*-
"""
gdcdatamodel.models.submission
----------------------------------

Models for submission TransactionLogs
"""

from datetime import datetime
from json import loads, dumps
from sqlalchemy import func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, deferred

import pytz

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    Index,
    Text,
    text,
)

Base = declarative_base()


def datetime_to_unix(dt):
    return (dt - datetime(1970, 1, 1, tzinfo=pytz.utc)).total_seconds()


class TransactionLog(Base):
    __tablename__ = "transaction_logs"

    @declared_attr
    def __table_args__(cls):
        tbl = cls.__tablename__
        return (
            Index("{}_program_idx".format(tbl), "program"),
            Index("{}_project_idx".format(tbl), "project"),
            Index("{}_is_dry_run_idx".format(tbl), "is_dry_run"),
            Index("{}_committed_by_idx".format(tbl), "committed_by"),
            Index("{}_closed_idx".format(tbl), "closed"),
            Index("{}_state_idx".format(tbl), "state"),
            Index("{}_submitter_idx".format(tbl), "submitter"),
            Index("{}_created_datetime_idx".format(tbl), "created_datetime"),
            Index("{}_project_id_idx".format(tbl), cls.program + "-" + cls.project),
        )

    def __repr__(self):
        return "<TransactionLog({}, {})>".format(self.id, self.created_datetime)

    def to_json(self, fields=set()):
        # Source fields
        existing_fields = [c.name for c in self.__table__.c] + ["entities", "documents"]
        custom_fields = {"created_datetime", "entities", "documents"}

        # Pull out child fields
        entity_fields = {f for f in fields if f.startswith("entities.")}
        document_fields = {f for f in fields if f.startswith("documents.")}
        fields = fields - entity_fields - document_fields

        # Reformat child fields
        entity_fields = {f.replace("entities.", "") for f in entity_fields}
        document_fields = {f.replace("documents.", "") for f in document_fields}

        # Default fields
        if not fields:
            fields = {"id", "submitter", "role", "program", "created_datetime"}

        # Check for field existence
        if set(fields) - set(existing_fields):
            raise RuntimeError(
                "Fields do not exist: {}".format(
                    ", ".join((set(fields) - set(existing_fields)))
                )
            )

        # Set standard fields
        doc = {key: getattr(self, key) for key in fields if key not in custom_fields}

        # Add custom fields
        if "entities" in fields or entity_fields:
            doc["entities"] = [n.to_json(entity_fields) for n in self.entities]
        if "documents" in fields or document_fields:
            doc["documents"] = [n.to_json(document_fields) for n in self.documents]
        if "created_datetime" in fields:
            doc["created_datetime"] = self.created_datetime.isoformat("T")

        return doc

    id = Column(
        Integer,
        primary_key=True,
    )

    submitter = Column(
        Text,
    )

    role = Column(
        Text,
        nullable=False,
    )

    program = Column(
        Text,
        nullable=False,
    )

    project = Column(
        Text,
        nullable=False,
    )

    #: Specifies a non-dry_run transaction that repeated this
    #: transaction in an attempt to write to the database
    committed_by = Column(
        Integer,
    )

    #: Was this transaction a dry_run (for validation)
    is_dry_run = Column(
        Boolean,
        nullable=False,
    )

    #: Has this transaction succeeded, errored, failed, etc.
    state = Column(
        Text,
        nullable=False,
    )

    closed = Column(
        Boolean,
        default=False,
        nullable=False,
    )

    @hybrid_property
    def project_id(self):
        return self.program + "-" + self.project

    @project_id.expression
    def project_id(cls):
        return func.concat(cls.program, "-", cls.project)

    created_datetime = Column(
        DateTime(timezone=True),
        nullable=False,
        server_default=text("now()"),
    )

    canonical_json = deferred(
        Column(
            JSONB,
            nullable=False,
            default=JSONB.NULL,
        )
    )


class TransactionSnapshot(Base):
    __tablename__ = "transaction_snapshots"

    def __repr__(self):
        return "<TransactionSnapshot({}, {})>".format(self.id, self.transaction_id)

    def to_json(self, fields=set()):
        fields = set(fields)
        existing_fields = [c.name for c in self.__table__.c]
        if not fields:
            fields = existing_fields
        if set(fields) - set(existing_fields):
            raise RuntimeError(
                "Entity fields do not exist: {}".format(
                    ", ".join((set(fields) - set(existing_fields)))
                )
            )
        doc = {key: getattr(self, key) for key in fields}
        return doc

    id = Column(
        Integer,
        primary_key=True,
    )

    entity_id = Column(
        Text,
        nullable=False,
        index=True,
    )

    transaction_id = Column(
        Integer,
        ForeignKey("transaction_logs.id"),
        index=True,
    )

    action = Column(
        Text,
        nullable=False,
    )

    old_props = Column(
        JSONB,
        nullable=False,
        default=JSONB.NULL,
    )

    new_props = Column(
        JSONB,
        nullable=False,
        default=JSONB.NULL,
    )

    transaction = relationship("TransactionLog", backref="entities")

    __table_args__ = (Index("snapshot_transaction_idx", "transaction_id"),)


class TransactionDocument(Base):
    __tablename__ = "transaction_documents"

    def to_json(self, fields=set()):
        # Source fields
        fields = set(fields)
        existing_fields = {c.name for c in self.__table__.c}

        # Default fields
        if not fields:
            fields = existing_fields

        # Check field existence
        if set(fields) - set(existing_fields):
            raise RuntimeError(
                "Entity fields do not exist: {}".format(
                    ", ".join(fields - existing_fields)
                )
            )

        # Generate doc
        doc = {key: getattr(self, key) for key in fields}
        return doc

    id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )

    transaction_id = Column(
        Integer,
        ForeignKey("transaction_logs.id"),
        primary_key=True,
    )

    name = Column(
        Text,
    )

    doc_format = Column(
        Text,
        nullable=False,
    )

    doc = deferred(
        Column(
            Text,
            nullable=False,
        )
    )

    response_json = deferred(
        Column(
            JSONB,
        )
    )

    transaction = relationship("TransactionLog", backref="documents")

    __table_args__ = (Index("document_transaction_idx", "transaction_id"),)

    @property
    def is_json(self):
        if self.doc_format.upper() != "JSON":
            return False
        else:
            return True

    @property
    def is_xml(self):
        if self.doc_format.upper() != "XML":
            return False
        else:
            return True

    @property
    def json(self):
        if not self.is_json:
            return None
        return loads(self.doc)

    @json.setter
    def json(self, doc):
        self.doc_format = "JSON"
        self.doc = dumps(doc)

    @property
    def xml(self):
        if not self.is_xml:
            return None
        return self.doc

    @xml.setter
    def xml(self, doc):
        self.doc_format = "XML"
        self.doc = doc



from setuptools import setup, find_packages
from subprocess import check_output


def get_version():
    # https://github.com/uc-cdis/dictionaryutils/pull/37#discussion_r257898408
    try:
        tag = check_output(
            ["git", "describe", "--tags", "--abbrev=0", "--match=[0-9]*"]
        )
        return tag.decode("utf-8").strip("\n")
    except Exception:
        raise RuntimeError(
            "The version number cannot be extracted from git tag in this source "
            "distribution; please either download the source from PyPI, or check out "
            "from GitHub and make sure that the git CLI is available."
        )


setup(
    name="gen3datamodel",
    version='3.0.3',
    packages=find_packages(),
    install_requires=[
        "pytz>=2016.4",
        "graphviz~=0.4",
        "jsonschema~=3.2",
        "psqlgraph~=3.0",
        "cdisutils",
        "python-dateutil~=2.4",
        "sqlalchemy~=1.3",
    ],
    package_data={
        "gdcdatamodel": [
            "xml_mappings/*.yaml",
        ]
    },
    entry_points={
        "console_scripts": ["gdc_postgres_admin=gdcdatamodel.gdc_postgres_admin:main"]
    },
)



from functools import wraps


def validate(*types, **kwargs):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)

        return f

    return decorator



from gdcdatamodel import models as md
from psqlgraph import Node, Edge, PsqlGraphDriver

import unittest

host = "localhost"
user = "test"
password = "test"
database = "automated_test"
g = PsqlGraphDriver(host, user, password, database)


class TestValidators(unittest.TestCase):
    @staticmethod
    def new_portion():
        portion = md.Portion(
            **{
                "node_id": "case1",
                "is_ffpe": False,
                "portion_number": "01",
                "project_id": "CGCI-BLGSP",
                "state": "validated",
                "submitter_id": "PORTION-1",
                "weight": 54.0,
            }
        )
        portion.acl = ["acl1"]
        portion.sysan.update({"key1": "val1"})
        return portion

    @staticmethod
    def new_analyte():
        return md.Analyte(
            **{
                "node_id": "analyte1",
                "analyte_type": "Repli-G (Qiagen) DNA",
                "analyte_type_id": "W",
                "project_id": "CGCI-BLGSP",
                "state": "validated",
                "submitter_id": "TCGA-AR-A1AR-01A-31W",
            }
        )

    def setUp(self):
        pass

    def tearDown(self):
        self._clear_tables()

    def _clear_tables(self):
        conn = g.engine.connect()
        conn.execute("commit")
        for table in Node().get_subclass_table_names():
            if table != Node.__tablename__:
                conn.execute("delete from {}".format(table))
        for table in Edge.get_subclass_table_names():
            if table != Edge.__tablename__:
                conn.execute("delete from {}".format(table))
        conn.execute("delete from versioned_nodes")
        conn.execute("delete from _voided_nodes")
        conn.execute("delete from _voided_edges")
        conn.close()

    def test_round_trip(self):
        with g.session_scope() as session:
            portion = self.new_portion()
            analyte = self.new_analyte()
            portion.analytes = [analyte]
            session.add(portion)

        with g.session_scope() as session:
            portion = g.nodes(md.Portion).one()
            v_node = md.VersionedNode.clone(portion)
            session.add(v_node)

        with g.session_scope():
            v_node = g.nodes(md.VersionedNode).one()

        self.assertEqual(v_node.properties["is_ffpe"], False)
        self.assertEqual(v_node.properties["state"], "validated")
        self.assertEqual(v_node.properties["state"], "validated")
        self.assertEqual(v_node.system_annotations, {"key1": "val1"})
        self.assertEqual(v_node.acl, ["acl1"])
        self.assertEqual(v_node.neighbors, ["analyte1"])
        self.assertIsNotNone(v_node.versioned)
        self.assertIsNotNone(v_node.key)

    def test_versions_property(self):
        with g.session_scope() as session:
            portion = self.new_portion()
            analyte = self.new_analyte()
            portion.analytes = [analyte]
            session.add(portion)

        with g.session_scope() as session:
            portion = g.nodes(md.Portion).one()
            v_node = md.VersionedNode.clone(portion)
            session.add(v_node)

        with g.session_scope():
            portion = g.nodes(md.Portion).one()
            portion._versions.one()

        with self.assertRaises(RuntimeError):
            portion._versions.one()

        with g.session_scope() as s:
            portion.get_versions(s).one()



from psqlgraph import Node, Edge

traversals = {}
terminal_nodes = [
    "annotations",
    "centers",
    "archives",
    "tissue_source_sites",
    "files",
    "related_files",
    "describing_files",
    "clinical_metadata_files",
    "experiment_metadata_files",
    "run_metadata_files",
    "analysis_metadata_files",
    "biospecimen_metadata_files",
    "aligned_reads_metrics",
    "read_group_metrics",
    "pathology_reports",
    "simple_germline_variations",
    "aligned_reads_indexes",
    "mirna_expressions",
    "exon_expressions",
    "simple_somatic_mutations",
    "gene_expressions",
    "aggregated_somatic_mutations",
]


def construct_traversals(root, node, visited, path):
    recurse = lambda neighbor: (
        neighbor
        # no backtracking
        and neighbor not in visited
        and neighbor != node
        # no traveling THROUGH terminal nodes
        and (
            path[-1] not in terminal_nodes
            if path
            else neighbor.label not in terminal_nodes
        )
        and (
            not path[-1].startswith("_related")
            if path
            else not neighbor.label.startswith("_related")
        )
    )

    for edge in Edge._get_edges_with_src(node.__name__):
        neighbor = [
            n for n in Node.get_subclasses() if n.__name__ == edge.__dst_class__
        ][0]
        if recurse(neighbor):
            construct_traversals(
                root, neighbor, visited + [node], path + [edge.__src_dst_assoc__]
            )

    for edge in Edge._get_edges_with_dst(node.__name__):
        neighbor = [
            n for n in Node.get_subclasses() if n.__name__ == edge.__src_class__
        ][0]
        if recurse(neighbor):
            construct_traversals(
                root, neighbor, visited + [node], path + [edge.__dst_src_assoc__]
            )

    traversals[root][node.label] = traversals[root].get(node.label) or set()
    traversals[root][node.label].add(".".join(path))


def construct_traversals_for_all_nodes():
    for node in Node.get_subclasses():
        traversals[node.label] = {}
        construct_traversals(node.label, node, [node], [])


def union_subq_without_path(q, *args, **kwargs):
    return q.except_(union_subq_path(q, *args, **kwargs))


def union_subq_path(q, dst_label, post_filters=[]):
    if traversals == {}:
        construct_traversals_for_all_nodes()
    src_label = q.entity().label
    if not traversals.get(src_label, {}).get(dst_label, {}):
        return q
    paths = list(traversals[src_label][dst_label])
    base = q.subq_path(paths.pop(), post_filters)
    while paths:
        base = base.union(q.subq_path(paths.pop(), post_filters))
    return base



import unittest
import uuid
from gdcdatamodel.validators import GDCJSONValidator, GDCGraphValidator
from psqlgraph import PsqlGraphDriver
from gdcdatamodel.models import *

host = "localhost"
user = "test"
password = "test"
database = "automated_test"
g = PsqlGraphDriver(host, user, password, database)


class MockSubmissionEntity(object):
    def __init__(self):
        self.errors = []
        self.node = None
        self.doc = {}

    def record_error(self, message, **kwargs):
        self.errors.append(dict(message=message, **kwargs))


class TestValidators(unittest.TestCase):
    def setUp(self):
        self.graph_validator = GDCGraphValidator()
        self.json_validator = GDCJSONValidator()
        self.entities = [MockSubmissionEntity()]

    def tearDown(self):
        self._clear_tables()

    def _clear_tables(self):
        conn = g.engine.connect()
        conn.execute("commit")
        for table in Node().get_subclass_table_names():
            if table != Node.__tablename__:
                conn.execute("delete from {}".format(table))
        for table in Edge.get_subclass_table_names():
            if table != Edge.__tablename__:
                conn.execute("delete from {}".format(table))
        conn.execute("delete from _voided_nodes")
        conn.execute("delete from _voided_edges")
        conn.close()

    def test_json_validator_with_insufficient_properties(self):
        self.entities[0].doc = {"type": "aliquot", "centers": {"submitter_id": "test"}}
        self.json_validator.record_errors(self.entities)
        self.assertEqual(self.entities[0].errors[0]["keys"], ["submitter_id"])
        self.assertEqual(1, len(self.entities[0].errors))

    def test_json_validator_with_wrong_node_type(self):
        self.entities[0].doc = {"type": "aliquo"}
        self.json_validator.record_errors(self.entities)
        self.assertEqual(self.entities[0].errors[0]["keys"], ["type"])
        self.assertEqual(1, len(self.entities[0].errors))

    def test_json_validator_with_wrong_property_type(self):
        self.entities[0].doc = {
            "type": "aliquot",
            "submitter_id": 1,
            "centers": {"submitter_id": "test"},
        }
        self.json_validator.record_errors(self.entities)
        self.assertEqual(["submitter_id"], self.entities[0].errors[0]["keys"])
        self.assertEqual(1, len(self.entities[0].errors))

    def test_json_validator_with_multiple_errors(self):
        self.entities[0].doc = {
            "type": "aliquot",
            "submitter_id": 1,
            "test": "test",
            "centers": {"submitter_id": "test"},
        }
        self.json_validator.record_errors(self.entities)
        self.assertEqual(2, len(self.entities[0].errors))

    def test_json_validator_with_nested_error_keys(self):
        self.entities[0].doc = {
            "type": "aliquot",
            "submitter_id": "test",
            "centers": {"submitter_id": True},
        }
        self.json_validator.record_errors(self.entities)
        self.assertEqual(["centers"], self.entities[0].errors[0]["keys"])

    def test_json_validator_with_multiple_entities(self):
        self.entities[0].doc = {
            "type": "aliquot",
            "submitter_id": 1,
            "test": "test",
            "centers": {"submitter_id": "test"},
        }
        entity = MockSubmissionEntity()
        entity.doc = {
            "type": "aliquot",
            "submitter_id": "test",
            "centers": {"submitter_id": "test"},
        }
        self.entities.append(entity)

        self.json_validator.record_errors(self.entities)
        self.assertEqual(2, len(self.entities[0].errors))
        self.assertEqual(0, len(entity.errors))

    def create_node(self, doc, session):
        cls = Node.get_subclass(doc["type"])
        node = cls(str(uuid.uuid4()))
        node.props = doc["props"]
        for key, value in doc["edges"].items():
            for target_id in value:
                edge = g.nodes().ids(target_id).first()
                node[key].append(edge)
        session.add(node)
        return node

    def update_schema(self, entity, key, schema):
        self.graph_validator.schemas.schema[entity][key] = schema

    def test_graph_validator_without_required_link(self):
        with g.session_scope() as session:
            node = self.create_node(
                {"type": "aliquot", "props": {"submitter_id": "test"}, "edges": {}},
                session,
            )
            self.entities[0].node = node
            self.update_schema(
                "aliquot",
                "links",
                [
                    {
                        "name": "analytes",
                        "backref": "aliquots",
                        "label": "derived_from",
                        "multiplicity": "many_to_one",
                        "target_type": "analyte",
                        "required": True,
                    }
                ],
            )
            self.graph_validator.record_errors(g, self.entities)
            self.assertEqual(["analytes"], self.entities[0].errors[0]["keys"])

    def test_graph_validator_with_exclusive_link(self):
        with g.session_scope() as session:
            analyte = self.create_node(
                {
                    "type": "analyte",
                    "props": {
                        "submitter_id": "test",
                        "analyte_type_id": "D",
                        "analyte_type": "DNA",
                    },
                    "edges": {},
                },
                session,
            )
            sample = self.create_node(
                {
                    "type": "sample",
                    "props": {
                        "submitter_id": "test",
                        "sample_type": "DNA",
                        "sample_type_id": "01",
                    },
                    "edges": {},
                },
                session,
            )

            node = self.create_node(
                {
                    "type": "aliquot",
                    "props": {"submitter_id": "test"},
                    "edges": {
                        "analytes": [analyte.node_id],
                        "samples": [sample.node_id],
                    },
                },
                session,
            )
            self.entities[0].node = node
            self.update_schema(
                "aliquot",
                "links",
                [
                    {
                        "exclusive": True,
                        "required": True,
                        "subgroup": [
                            {
                                "name": "analytes",
                                "backref": "aliquots",
                                "label": "derived_from",
                                "multiplicity": "many_to_one",
                                "target_type": "analyte",
                            },
                            {
                                "name": "samples",
                                "backref": "aliquots",
                                "label": "derived_from",
                                "multiplicity": "many_to_one",
                                "target_type": "sample",
                            },
                        ],
                    }
                ],
            )
            self.graph_validator.record_errors(g, self.entities)
            self.assertEqual(
                ["analytes", "samples"], self.entities[0].errors[0]["keys"]
            )

    def test_graph_validator_with_wrong_multiplicity(self):
        with g.session_scope() as session:
            analyte = self.create_node(
                {
                    "type": "analyte",
                    "props": {
                        "submitter_id": "test",
                        "analyte_type_id": "D",
                        "analyte_type": "DNA",
                    },
                    "edges": {},
                },
                session,
            )

            analyte_b = self.create_node(
                {
                    "type": "analyte",
                    "props": {
                        "submitter_id": "testb",
                        "analyte_type_id": "H",
                        "analyte_type": "RNA",
                    },
                    "edges": {},
                },
                session,
            )

            node = self.create_node(
                {
                    "type": "aliquot",
                    "props": {"submitter_id": "test"},
                    "edges": {"analytes": [analyte.node_id, analyte_b.node_id]},
                },
                session,
            )
            self.entities[0].node = node
            self.update_schema(
                "aliquot",
                "links",
                [
                    {
                        "exclusive": False,
                        "required": True,
                        "subgroup": [
                            {
                                "name": "analytes",
                                "backref": "aliquots",
                                "label": "derived_from",
                                "multiplicity": "many_to_one",
                                "target_type": "analyte",
                            },
                            {
                                "name": "samples",
                                "backref": "aliquots",
                                "label": "derived_from",
                                "multiplicity": "many_to_one",
                                "target_type": "sample",
                            },
                        ],
                    }
                ],
            )
            self.graph_validator.record_errors(g, self.entities)
            self.assertEqual(["analytes"], self.entities[0].errors[0]["keys"])

    def test_graph_validator_with_correct_node(self):
        with g.session_scope() as session:
            analyte = self.create_node(
                {
                    "type": "analyte",
                    "props": {
                        "submitter_id": "test",
                        "analyte_type_id": "D",
                        "analyte_type": "DNA",
                    },
                    "edges": {},
                },
                session,
            )

            node = self.create_node(
                {
                    "type": "aliquot",
                    "props": {"submitter_id": "test"},
                    "edges": {"analytes": [analyte.node_id]},
                },
                session,
            )
            self.entities[0].node = node
            self.update_schema(
                "aliquot",
                "links",
                [
                    {
                        "exclusive": False,
                        "required": True,
                        "subgroup": [
                            {
                                "name": "analytes",
                                "backref": "aliquots",
                                "label": "derived_from",
                                "multiplicity": "many_to_one",
                                "target_type": "analyte",
                            },
                            {
                                "name": "samples",
                                "backref": "aliquots",
                                "label": "derived_from",
                                "multiplicity": "many_to_one",
                                "target_type": "sample",
                            },
                        ],
                    }
                ],
            )
            self.graph_validator.record_errors(g, self.entities)
            self.assertEqual(0, len(self.entities[0].errors))
